@@ -2,14 +2,13 @@
 Variogram class
 """
 import copy
-import os
 import warnings
 
 import numpy as np
 from pandas import DataFrame
-import matplotlib.pyplot as plt
-from scipy.optimize import curve_fit
+from scipy.optimize import curve_fit, minimize, OptimizeWarning
 from scipy.spatial.distance import pdist, squareform
+from scipy import stats 
 from sklearn.isotonic import IsotonicRegression
 
 from skgstat import estimators, models, binning
@@ -92,17 +91,35 @@ class Variogram(object):
             scipy.spatial.distance.pdist. Additional parameters are not (yet)
             passed through to pdist. These are accepted by pdist for some of
             the metrics. In these cases the default values are used.
-        bin_func : str
+        bin_func : str            
+            .. versionchanged:: 0.3.8
+                added 'fd', 'sturges', 'scott', 'sqrt', 'doane'
+            .. versionchanged:: 0.3.9
+                added 'kmeans', 'ward'
+
             String identifying the binning function used to find lag class
-            edges. At the moment there are two possible values: 'even'
-            (default) or 'uniform'. Even will find n_lags bins of same width
-            in the interval [0,maxlag[. 'uniform' will identfy n_lags bins on
-            the same interval, but with varying edges so that all bins count
-            the same amount of observations.
+            edges. All methods calculate bin edges on the interval [0, maxlag[.
+            Possible values are:
+
+                * `'even'` (default) finds `n_lags` same width bins
+                * `'uniform'` forms `n_lags` bins of same data count
+                * `'fd'` applies Freedman-Diaconis estimator to find `n_lags`
+                * `'sturges'` applies Sturge's rule to find `n_lags`.
+                * `'scott'` applies Scott's rule to find `n_lags`
+                * `'doane'` applies Doane's extension to Sturge's rule to find `n_lags`
+                * `'sqrt'` uses the square-root of :func:`distance <skgstat.Variogram.distance>` as `n_lags`.
+                * `'kmeans'` uses KMeans clustering to well supported bins
+                * `'ward'` uses hierachical clustering to find minimum-variance clusters.
+
+            More details are given in the documentation for :func:`set_bin_func <skgstat.Variogram.set_bin_func>`.
+
         normalize : bool
             Defaults to False. If True, the independent and dependent
             variable will be normalized to the range [0,1].
         fit_method : str
+            .. versionchanged:: 0.3.10
+                Added 'ml' and 'custom'
+
             String identifying the method to be used for fitting the
             theoretical variogram function to the experimental. More info is
             given in the Variogram.fit docs. Can be one of:
@@ -113,6 +130,15 @@ class Variogram(object):
                 * 'trf': Trust Region Reflective function for non-linear
                   constrained problems. The class will set the boundaries
                   itself. This is the default function.
+                * 'ml': Maximum-Likelihood estimation. With the current implementation
+                  only the Nelder-Mead solver for unconstrained problems is
+                  implemented. This will estimate the variogram parameters from
+                  a Gaussian parameter space by minimizing the negative
+                  log-likelihood.
+                * 'manual': Manual fitting. You can set the range, sill and
+                  nugget either directly to the :func:`fit <skgstat.Variogram.fit>`
+                  function, or as `fit_` prefixed keyword arguments on
+                  Variogram instantiation.
 
         fit_sigma : numpy.ndarray, str
             Defaults to None. The sigma is used as measure of uncertainty
@@ -156,13 +182,29 @@ class Variogram(object):
         -----------------
         entropy_bins : int, str
             .. versionadded:: 0.3.7
+
             If the `estimator <skgstat.Variogram.estimator>` is set to
             `'entropy'` this argument sets the number of bins, that should be
             used for histogram calculation.
         percentile : int
             .. versionadded:: 0.3.7
+
             If the `estimator <skgstat.Variogram.estimator>` is set to 
             `'entropy'` this argument sets the percentile to be used.
+        binning_random_state : int, None
+            .. versionadded:: 0.3.9
+
+            If :func:`bin_func <skgstat.Variogram.set_bin_func>` is `'kmeans'`
+            this can overwrite the seed for the initial guess of the cluster
+            centroids. Note, that K-Means is not deterministic and is therefore
+            seeded to 42 here. You can pass `None` to disable this behavior,
+            but use it with care, as you will get different results.
+        binning_agg_func : str
+            .. versionadded:: 0.3.10
+
+            If :func:`bin_func <skgstat.Variogram.set_bin_func>` is `'ward'`
+            this keyword argument can switch from default mean aggregation to
+            median aggregation for calculating the cluster centroids.
 
         """
         # Before we do anything else, make kwargs available
@@ -378,32 +420,113 @@ class Variogram(object):
         self.set_bin_func(bin_func=bin_func)
 
     def set_bin_func(self, bin_func: str):
-        """Set binning function
+        r"""Set binning function
 
         Sets a new binning function to be used. The new binning method is set
         by a string identifying the new function to be used. Can be one of:
-        ['even', 'uniform'].
+        ['even', 'uniform', 'fd', 'sturges', 'scott', 'sqrt', 'doane'].
+        If the number of lag classes should be estimated automatically, it is 
+        recommended to use ' sturges' for small, normal distributed locations
+        and 'fd' or 'scott' for large datasets, where 'fd' is more robust to
+        outliers. 'sqrt' is by far the fastest estimator. 'doane' is an 
+        extension of Sturge's rule for non-normal distributed data.
+
+        .. versionchanged:: 0.3.8
+            added 'fd', 'sturges', 'scott', 'sqrt', 'doane'
+        
+        .. versionchanged:: 0.3.9
+            added 'kmeans', 'ward'
 
         Parameters
         ----------
         bin_func : str
             Can be one of:
 
-            * **'even'**: Use skgstat.binning.even_width_lags for using
-              n_lags lags of equal width up to maxlag.
-            * **'uniform'**: Use skgstat.binning.uniform_count_lags for using
-              n_lags lags up to maxlag in which the pairwise differences
-              follow a uniform distribution.
+                * 'even'
+                * 'uniform'
+                * 'fd'
+                * 'sturges'
+                * 'scott'
+                * 'sqrt'
+                * 'doane'
+                * 'kmeans'
+                * 'ward'
 
         Returns
         -------
         void
 
+        Notes
+        -----
+        **`'even'`**: Use skgstat.binning.even_width_lags for using
+        n_lags lags of equal width up to maxlag.
+
+        **`'uniform'`**: Use skgstat.binning.uniform_count_lags for using
+        n_lags lags up to maxlag in which the pairwise differences
+        follow a uniform distribution.
+
+        **`'sturges'`**: estimates the number of evenly distributed lag
+        classes (n) by Sturges rule [101]_:
+
+        .. math::
+            n = log_2 n + 1
+
+        **`'scott'`**: estimates the lag class widths (h) by Scott's rule [102]_:
+
+        .. math::
+            h = \sigma \frac{24 * \sqrt{\pi}}{n}^{\frac{1}{3}}
+
+        **`'sqrt'`**: estimates the number of lags (n) by the suare-root:
+
+        .. math::
+            n = \sqrt{n}
+
+        **`'fd'`**: estimates the lag class widths (h) using the
+        Freedman Diaconis estimator [103]_:
+
+        .. math::
+            h = 2\frac{IQR}{n^{1/3}}
+
+        **`'doane'`**: estimates the number of evenly distributed lag classes
+        using Doane's extension to Sturge's rule [104]_:
+
+        .. math::
+            n = 1 + \log_{2}(s) + \log_2\left(1 + \frac{|g|}{k}\right)
+            g = E\left[\left(\frac{x - \mu_g}{\sigma}\right)^3\right]
+            k = \sqrt{\frac{6(s - 2)}{(s + 1)(s + 3)}}
+
+        **`'kmeans'`**: This method will search for `n` clusters in the
+        distance matrix. The cluster centroids are used to calculate the
+        upper edges of the lag classes, by setting it to half of the distance
+        between two neighboring clusters. Note: This does not necessarily
+        result in even width bins.
+
+        **`'ward'`** uses a hierachical culstering algorithm to iteratively
+        merge pairs of clusters until there are only `n` remaining clusters.
+        The merging is done by minimizing the variance for the merged cluster.
+
         See Also
         --------
         Variogram.bin_func
         skgstat.binning.uniform_count_lags
         skgstat.binning.even_width_lags
+        skgstat.binning.auto_derived_lags
+        skgstat.binning.kmeans
+        skgstat.binning.ward
+        sklearn.cluster.KMeans
+        sklearn.cluster.AgglomerativeClustering
+
+        References
+        ----------
+        .. [101] Scott, D.W. (2009), Sturges' rule. WIREs Comp Stat, 1: 303-306. 
+            https://doi.org/10.1002/wics.35
+        .. [102] Scott, D.W. (2010), Scott's rule. WIREs Comp Stat, 2: 497-502. 
+            https://doi.org/10.1002/wics.103
+        .. [103] Freedman, David, and Persi Diaconis  (1981), "On the histogram as 
+            a density estimator: L 2 theory." Zeitschrift für Wahrscheinlichkeitstheorie 
+            und verwandte Gebiete 57.4: 453-476.
+        .. [104] Doane, D. P. (1976). Aesthetic frequency classifications. 
+            The American Statistician, 30(4), 181-183.
 
         """
         # switch the input
@@ -411,9 +534,26 @@ class Variogram(object):
             self._bin_func = binning.even_width_lags
         elif bin_func.lower() == 'uniform':
             self._bin_func = binning.uniform_count_lags
+        elif bin_func.lower() == 'kmeans':
+            # define a helper to pass kwargs
+            def wrapper(distance, n, maxlag):
+                return binning.kmeans(distance, n, maxlag, **self._kwargs)
+            self._bin_func = wrapper
+        elif bin_func.lower() == 'ward':
+            self._bin_func = binning.ward
+        elif isinstance(bin_func, str):
+            # define a helper wrapper
+            def wrapper(distances, n, maxlag):
+                return binning.auto_derived_lags(distances, bin_func.lower(), maxlag)
+
+            self._bin_func = wrapper
+            self._n_lags = None
+        elif callable(bin_func):
+            self._bin_func = bin_func
+            bin_func = 'custom'
         else:
-            raise ValueError('%s binning method is not known' % bin_func)
-        
+            raise AttributeError('bin_func has to be of type string.')
+
         # store the name
         self._bin_func_name = bin_func
 
@@ -433,8 +573,14 @@ class Variogram(object):
 
     @property
     def bins(self):
+        # if bins are not calculated, do it
         if self._bins is None:
-            self._bins = self.bin_func(self.distance, self.n_lags, self.maxlag)
+            self._bins, n = self.bin_func(self.distance, self._n_lags, self.maxlag)
+
+            # if the binning function returned an N, the n_lags need
+            # to be adjusted directly (not through the setter)
+            if n is not None:
+                self._n_lags = n
 
         return self._bins.copy()
 
@@ -457,6 +603,8 @@ class Variogram(object):
         the grouping index and fitting parameters
 
         """
+        if self._n_lags is None:
+            self._n_lags = len(self.bins)
         return self._n_lags
 
     @n_lags.setter
@@ -915,6 +1063,9 @@ class Variogram(object):
         overwritten. All other keyword arguments will be passed to
         scipy.optimize.curve_fit function.
 
+        .. versionchanged:: 0.3.10
+            added 'ml' and 'custom' method.
+
         Parameters
         ----------
         force : bool
@@ -923,12 +1074,22 @@ class Variogram(object):
             False.
         method : string
             A string identifying one of the implemented fitting procedures.
-            Can be one of ['lm', 'trf']:
+            Can be one of:
 
               * lm: Levenberg-Marquardt algorithms implemented in
                 scipy.optimize.leastsq function.
               * trf: Trust Region Reflective algorithm implemented in
                 scipy.optimize.least_squares(method='trf')
+              * 'ml': Maximum-Likelihood estimation. With the current implementation
+                only the Nelder-Mead solver for unconstrained problems is
+                implemented. This will estimate the variogram parameters from
+                a Gaussian parameter space by minimizing the negative
+                log-likelihood.
+              * 'manual': Manual fitting. You can set the range, sill and
+                nugget either directly to the :func:`fit <skgstat.Variogram.fit>`
+                function, or as `fit_` prefixed keyword arguments on
+                Variogram instantiation.
+
 
         sigma : string, array
             Uncertainty array for the bins. Has to have the same dimension as
@@ -940,13 +1101,18 @@ class Variogram(object):
 
         See Also
         --------
-        scipy.optimize
+        scipy.optimize.minimize
         scipy.optimize.curve_fit
         scipy.optimize.leastsq
         scipy.optimize.least_squares
 
         """
-        # TODO: the kwargs need to be preserved somehow
+        # store the old cof
+        if self.cof is None:
+            old_params = {}
+        else:
+            old_params = self.describe()
+
         # delete the last cov and cof
         self.cof = None
         self.cov = None
@@ -984,17 +1150,27 @@ class Variogram(object):
             self.cof = [r, s, n]
             return
 
-
         # Switch the method
+        # wrap the model to include or exclude the nugget
+        if self.use_nugget:
+            def wrapped(*args):
+                return self._model(*args)
+        else:
+            def wrapped(*args):
+                return self._model(*args, 0)
+
+        # get p0
+        bounds = (0, self.__get_fit_bounds(x, y))
+        p0 = np.asarray(bounds[1])
+
         # Trust Region Reflective
         if self.fit_method == 'trf':
-            bounds = (0, self.__get_fit_bounds(x, y))
             self.cof, self.cov = curve_fit(
-                self._model,
+                wrapped,
                 _x, _y,
                 method='trf',
                 sigma=self.fit_sigma,
-                p0=bounds[1],
+                p0=p0,
                 bounds=bounds,
                 **kwargs
             )
@@ -1002,15 +1178,72 @@ class Variogram(object):
         # Levenberg-Marquardt
         elif self.fit_method == 'lm':
             self.cof, self.cov = curve_fit(
-                self.model,
+                wrapped,
                 _x, _y,
                 method='lm',
                 sigma=self.fit_sigma,
+                p0=p0,
                 **kwargs
             )
 
+        # maximum-likelihood
+        elif self.fit_method == 'ml':
+            # check if the probabilities must be weighted
+            if self.fit_sigma is None:
+                sigma = np.ones(self.bins.size)
+            else:
+                sigma = 1 / self.fit_sigma
+
+            # define the loss function to be minimized
+            def ml(params):
+                # predict
+                pred = [wrapped(_, *params) for _ in _x]
+
+                # get the probabilities of _y
+                p = [stats.norm.logpdf(_p, loc=o, scale=1.) for _p, o in zip(pred, _y)]
+
+                # weight the probs
+                return - np.sum(p * sigma)
+
+            # apply maximum likelihood estimation by minimizing ml
+            result = minimize(ml, p0 * 0.5, method='SLSQP', bounds=[(0, _) for _ in p0])
+
+            if not result.success:  # pragma: no cover
+                raise OptimizeWarning('Maximum Likelihood could not estimate parameters.')
+            else:
+                # set the result
+                self.cof = result.x
+
+        # manual fitting
+        elif self.fit_method == 'manual':
+            # TODO: here, the Error could only be raises if cof was None so far
+            r = kwargs.get('range', self._kwargs.get('fit_range', old_params.get('effective_range')))
+            s = kwargs.get('sill', self._kwargs.get('fit_sill', old_params.get('sill')))
+
+            # if not given raise an AttributeError
+            if r is None or s is None:
+                raise AttributeError('For manual fitting, you need to pass the \
+                    variogram parameters either to fit or to the Variogram \
+                    instance.\n parameter need to be prefixed with fit_ if \
+                    passed to __init__.')
+
+            # get the nugget
+            n = kwargs.get('nugget', self._kwargs.get('fit_nugget', old_params.get('nugget', 0.0)))
+
+            # check if a s parameter is needed
+            if self._model.__name__ in ('stable', 'matern'):
+                if self._model.__name__ == 'stable':
+                    s2 = kwargs.get('shape', self._kwargs.get('fit_shape', old_params.get('shape',2.0)))
+                if self._model.__name__ == 'matern':
+                    s2 = kwargs.get('shape', self._kwargs.get('fit_shape', old_params.get('smoothness', 2.0)))
+
+                # set
+                self.cof = [r, s, s2, n]
+            else:
+                self.cof = [r, s, n]
+
         else:
-            raise ValueError("fit method has to be one of ['trf', 'lm']")
+            raise ValueError("fit method has to be one of ['trf', 'lm', 'ml', 'custom']")
 
     def transform(self, x):
         """Transform
@@ -1247,7 +1480,7 @@ class Variogram(object):
 
         # if use_nugget is True add the nugget
         if self.use_nugget:
-            bounds.append(0.99)
+            bounds.append(0.99*np.nanmax(y))
 
         return bounds
 
@@ -1757,16 +1990,21 @@ class Variogram(object):
         -----------------
         add_trend_line : bool
             .. versionadded:: 0.3.5
+
             If set to `True`, the class will fit a linear model to each
             coordinate dimension and output the model along with a
             calculated R². With high R² values, you should consider
             rejecting the input data, or transforming it.
+
             .. note::
                 Right now, this is only supported for ``'plotly'`` backend
+   
 
         Returns
         -------
-        matplotlib.Figure, plotly.graph_objects.Figure
+        fig : matplotlib.Figure, plotly.graph_objects.Figure
+            The figure produced by the function. Dependends on the 
+            current backend.
 
         """
         # get the backend
