@@ -53,13 +53,13 @@ class TestVariogramInstatiation(unittest.TestCase):
 
         for b, e in zip(bins, V.bins):
             self.assertAlmostEqual(b, e, places=2)
-    
+
     def test_unknown_binning_func(self):
         with self.assertRaises(ValueError) as e:
             Variogram(self.c, self.v, bin_func='notafunc')
 
         self.assertEqual(
-            'notafunc binning method is not known',
+            "'notafunc' is not a valid estimator for `bins`",
             str(e.exception)
         )
 
@@ -117,6 +117,29 @@ class TestVariogramArguments(unittest.TestCase):
         V.bin_func = 'even'
         assert_array_almost_equal(even, V.bins, decimal=2)
 
+    def test_binning_method_scott(self):
+        V = Variogram(self.c, self.v, bin_func='scott')
+
+        # scott should yield 11 bins here
+        self.assertTrue(V.n_lags == 11)
+
+        assert_array_almost_equal(
+            V.bins,
+            np.array([4.9, 8.6, 12.4, 16.1, 19.9, 23.6, 27.3, 31.1, 34.8, 38.6, 42.3]),
+            decimal=1
+        )
+
+    def test_binning_change_nlags(self):
+        V = Variogram(self.c, self.v, n_lags=5)
+
+        # 5 lags are awaited
+        self.assertTrue(V.n_lags == 5)
+
+        # switch to fd rule
+        V.bin_func = 'fd'
+
+        self.assertTrue(V.n_lags == 13)
+
     def test_set_bins_directly(self):
         V = Variogram(self.c, self.v, n_lags=5)
 
@@ -131,6 +154,36 @@ class TestVariogramArguments(unittest.TestCase):
         self.assertIsNone(V.cov)
         self.assertIsNone(V.cof)
 
+    def test_binning_non_string_arg(self):
+        V = Variogram(self.c, self.v, n_lags=8)
+
+        return True
+
+    def test_binning_kmeans_method(self):
+        V = Variogram(
+            self.c,
+            self.v,
+            n_lags=6,
+            bin_func='kmeans',
+            binning_random_state=1306
+        )
+
+        assert_array_almost_equal(
+            V.bins,
+            np.array([2.5, 7.7, 12.9, 18.1, 23.7, 30.3]),
+            decimal=1
+        )
+
+    def test_binning_ward_method(self):
+        V = Variogram(self.c, self.v, n_lags=6, bin_func='ward')
+
+        assert_array_almost_equal(
+            V.bins,
+            np.array([2.5,  7.1, 11.1, 16.2, 23., 30.]),
+            decimal=1
+        )
+
+
     def test_estimator_method_setting(self):
         """
         Only test if the estimator functions are correctly set. The
@@ -363,13 +416,6 @@ class TestVariogramFittingProcedure(unittest.TestCase):
         for x, y in zip(sigs, self.V.fit_sigma):
             self.assertEqual(x, y)
 
-        # test parameter estimated
-        self.V.fit()
-        assert_array_almost_equal(
-            self.V.parameters,
-            [24.008, 17.083, 0.99], decimal=3
-        )
-
     def test_fit_sigma_raises_AttributeError(self):
         self.V.fit_sigma = (0, 1, 2)
 
@@ -403,7 +449,7 @@ class TestVariogramFittingProcedure(unittest.TestCase):
         # test parameters:
         self.V.fit()
         assert_array_almost_equal(
-            self.V.parameters, [25.077, 17.393, 0.925], decimal=3
+            self.V.parameters, [13., 0.3, 18.], decimal=1
         )
 
     def test_fit_sigma_exp(self):
@@ -416,7 +462,7 @@ class TestVariogramFittingProcedure(unittest.TestCase):
 
         # test parameters
         assert_array_almost_equal(
-            self.V.parameters, [25.3, 17.7, 1.], decimal=1
+            self.V.parameters, [25., 0.2, 18.5], decimal=1
         )
 
     def test_fit_sigma_sqrt(self):
@@ -429,7 +475,7 @@ class TestVariogramFittingProcedure(unittest.TestCase):
 
         # test the parameters
         assert_array_almost_equal(
-            self.V.parameters, [23., 17.,  1.], decimal=1
+            self.V.parameters, [19.7, 1.5,  16.4], decimal=1
         )
 
     def test_fit_sigma_sq(self):
@@ -442,7 +488,7 @@ class TestVariogramFittingProcedure(unittest.TestCase):
 
         # test the parameters
         assert_array_almost_equal(
-            self.V.parameters, [25.3, 17.6,  1.], decimal=1
+            self.V.parameters, [5.4, 0.1,  18.5], decimal=1
         )
     
     def test_fit_sigma_on_the_fly(self):
@@ -455,21 +501,29 @@ class TestVariogramFittingProcedure(unittest.TestCase):
 
         # test the parameters
         assert_array_almost_equal(
-            self.V.parameters, [25.3, 17.6,  1.], decimal=1
+            self.V.parameters, [5.4, 0.1,  18.5], decimal=1
         )
 
     def test_fit_lm(self):
-        self.V.fit(method='lm', sigma='sqrt')
+        df = pd.read_csv(os.path.dirname(__file__) + '/sample.csv')
+        V = Variogram(
+            df[['x', 'y']],
+            df.z.values,
+            use_nugget=True,
+            n_lags=8, fit_method='lm'
+        )
 
         # test the parameters
         assert_array_almost_equal(
-            self.V.parameters, [1., 17., 1.], decimal=0
+            V.parameters, [162.3, 0.5, 0.8], decimal=1
         )
 
     def test_fitted_model(self):
+        self.V.fit_method = 'trf'
+        self.V.fit_sigma = None
         fun = self.V.fitted_model
 
-        result = [0.99, 7.19, 12.53, 16.14]
+        result = np.array([12.48, 17.2, 17.2, 17.2])
 
         assert_array_almost_equal(
             result, list(map(fun, np.arange(0, 20, 5))),
@@ -480,15 +534,14 @@ class TestVariogramFittingProcedure(unittest.TestCase):
         with self.assertRaises(ValueError) as e:
             self.V.fit(method='unsupported')
 
-        self.assertEqual(
-            "fit method has to be one of ['trf', 'lm']",
-            str(e.exception)
+        self.assertTrue(
+            "fit method has to be one of" in str(e.exception)
         )
-  
+
     def test_implicit_run_fit_fitted_model(self):
         self.V.fit_sigma = None
         self.V.fit_method = 'trf'
-        result = [0.99,  7.19, 12.53, 16.14]
+        result = np.array([12.48, 17.2, 17.2, 17.2])
 
         # remove cof
         self.V.cof = None
@@ -503,7 +556,7 @@ class TestVariogramFittingProcedure(unittest.TestCase):
     def test_implicit_run_fit_transform(self):
         self.V.fit_sigma = None
         self.V.fit_method = 'trf'
-        result = [0.99,  7.19, 12.53, 16.14]
+        result = np.array([12.48, 17.2, 17.2, 17.2])
 
         # test on transform
         self.V.cof = None
@@ -525,6 +578,85 @@ class TestVariogramFittingProcedure(unittest.TestCase):
             decimal=2
         )
 
+    def test_ml_default(self):
+        # load data sample
+        df = pd.read_csv(os.path.dirname(__file__) + '/sample.csv')
+        V = Variogram(
+            df[['x', 'y']],
+            df.z.values,
+            use_nugget=True,
+            n_lags=15,
+            fit_method='ml'
+        )
+
+        assert_array_almost_equal(
+            V.parameters, np.array([41.18, 1.2, 0.]), decimal=2
+        )
+
+    def test_ml_sq_sigma(self):
+        # load data sample
+        df = pd.read_csv(os.path.dirname(__file__) + '/sample.csv')
+        V = Variogram(
+            df[['x', 'y']],
+            df.z.values,
+            use_nugget=True,
+            n_lags=15,
+            fit_method='ml',
+            fit_sigma='sq'
+        )
+
+        assert_array_almost_equal(
+            V.parameters, np.array([42.72, 1.21, 0.]), decimal=2
+        )
+
+    def test_manual_fit(self):
+        V = Variogram(
+            self.c,
+            self.v,
+            fit_method='manual',
+            model='spherical',
+            fit_range=10.,
+            fit_sill=5.
+        )
+
+        self.assertEqual(V.parameters, [10., 5., 0.0])
+    
+    def test_manual_fit_change(self):
+        V = Variogram(
+            self.c,
+            self.v,
+            fit_method='trf',
+            model='matern',
+        )
+
+        # switch to manual fit
+        V.fit_method = 'manual'
+        V.fit(range=10, sill=5, shape=3)
+
+        self.assertEqual(V.parameters, [10., 5., 3., 0.0])
+
+    def test_manual_raises_missing_params(self):
+        with self.assertRaises(AttributeError) as e:
+            Variogram(self.c, self.v, fit_method='manual')
+            self.assertTrue('For manual fitting' in str(e.exception))
+
+    def test_manual_preserve_params(self):
+        V = Variogram(self.c, self.v, fit_method='trf', n_lags=8)
+        params = V.parameters
+
+        # switch fit method
+        V.fit_method = 'manual'
+        V.fit(sill=14)
+
+        # expected output
+        params[1] = 14.
+
+        assert_array_almost_equal(
+            V.parameters,
+            params,
+            decimal=1
+        )      
+
 
 class TestVariogramQaulityMeasures(unittest.TestCase):
     def setUp(self):
